"""
Copyright (C) 2025 drd <drd.ltt000@gmail.com>

This file is part of TunEd.

TunEd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TunEd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys

if not __package__:
    # Make CLI runnable from source tree with
    #    python src/package
    package_source_path = os.path.dirname(os.path.dirname(__file__))
    sys.path.insert(0, package_source_path)
    
if __name__ == "__main__":
    from tuned.tuned import tuned
    
    tuned()
    


"""
Copyright (C) 2025 drd <drd.ltt000@gmail.com>

This file is part of TunEd.

TunEd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TunEd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import copy
import sys
import time
from datetime import timedelta
from threading import Thread

import sounddevice  # for silence PyAudio()

import numpy as np
from pyaudio import PyAudio, paInt16


class AudioAnalyzer(Thread):

    ZERO_PADDING = 3  # times the buffer length
    NUM_HPS = 3  # Harmonic Product Spectrum

    def __init__(self,
                 queue,
                 ref_freq=440,
                 detection='note',
                 sampling_rate=48000,
                 chunk_size=1024,
                 buffer_times=50,
                 *args,
                 **kwargs):
        Thread.__init__(self, *args, **kwargs)

        self.queue = queue
        self.detection = detection
        self.ref_freq = ref_freq
        # https://en.wikipedia.org/wiki/Sampling_(signal_processing)
        self.SAMPLING_RATE = sampling_rate  # 44100, 48000, 96000
        self.CHUNK_SIZE = chunk_size  # 256  # 512  # 1024  # number of samples
        self.BUFFER_TIMES = buffer_times  # 12  # 25  # 50  # buffer length = CHUNK_SIZE * BUFFER_TIMES

        self.buffer = np.zeros(self.CHUNK_SIZE * self.BUFFER_TIMES)
        self.hanning_window = np.hanning(len(self.buffer))
        # self.hamming_window = np.hamming(len(self.buffer))
        self.running = False

        try:
            self.audio_object = PyAudio()
            self.stream = self.audio_object.open(
                format=paInt16,
                channels=1,
                rate=self.SAMPLING_RATE,
                input=True,
                output=False,
                frames_per_buffer=self.CHUNK_SIZE
            )
        except Exception as e:
            sys.stderr.write('Error: Line {} {} {}\n'.format(sys.exc_info()[-1].tb_lineno, type(e).__name__, e))
            return

    @staticmethod
    def frequency_to_midi(frequency, ref_freq):
        """
        Converts a frequency into a MIDI note.
        Rounds to the closest midi note.

        :param frequency:
        :param ref_freq:
        :return:
        """
        if frequency == 0:
            return 0

        midi_note = int(round(69 + (12 * np.log2(frequency / ref_freq)) / np.log2(2)))

        return midi_note

    @staticmethod
    def midi_to_frequency(midi_note, ref_freq):
        """
        Converts a MIDI note into frequency.

        :param midi_note:
        :param ref_freq:
        :return:
        """
        frequency = ref_freq * 2.0**((midi_note - 69) / 12.0)

        return frequency

    @staticmethod
    def midi_to_ansi_note(midi_note):
        """
        Returns the Ansi Note name for a midi number.

        :param midi_note:
        :return:
        """
        if midi_note == 0:
            return "-", 0

        notes = ["A", "A#", "B", "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#"]
        num_notes = len(notes)

        note_name = notes[int((midi_note - 21) % num_notes)]
        note_number = (midi_note - 12) // num_notes

        return note_name, note_number

    # TODO
    @staticmethod
    def ainsi_note_to_midi(ainsi_note):
        """

        :param ainsi_note:
        :return:
        """
        pass

    def compute_frequency_offset(self, frequency, midi_note):
        """
        Compute offset in cent.

        :param frequency:
        :param midi_note:
        :return:
        """
        nearest_midi_note_frequency = self.midi_to_frequency(midi_note, self.ref_freq)
        frequency_offset = nearest_midi_note_frequency - frequency

        next_note = midi_note
        if frequency_offset > 0:
            next_note += 1
        elif frequency_offset < 0:
            next_note -= 1
        semitone_step = abs((nearest_midi_note_frequency - self.midi_to_frequency(next_note, self.ref_freq)) / 100)

        offset = round(frequency_offset/semitone_step)

        return offset

    def note_detection(self, magnitude_data, frequencies):
        # put the frequency of the loudest tone into the queue
        magnitude = np.max(magnitude_data)
        magnitude_to_db = 10 * np.log10(magnitude / 32768)  # reference value used for dBFS scale. 32768 for int16 and 1 for float
        index_loudest = np.argmax(magnitude_data)
        frequency = round(frequencies[index_loudest], 2)
        phase = np.angle(magnitude)
        midi_note = self.frequency_to_midi(frequency, self.ref_freq)
        note, octave = self.midi_to_ansi_note(midi_note)
        offset = self.compute_frequency_offset(frequency, midi_note)

        sound = {
            'magnitude': magnitude,
            'magnitude_to_db': 0 if np.isnan(magnitude_to_db) else magnitude_to_db,
            'phase': phase,
            'frequency': frequency,
            'note': note,
            'octave': octave,
            'offset': offset
        }

        return sound

    def run(self):

        self.running = True

        while self.running:
            try:
                # execution time
                start_time = time.perf_counter()

                # read microphone data
                frame = self.stream.read(self.CHUNK_SIZE)
                decoded_frame = np.frombuffer(frame, dtype=np.int16)

                # append data to audio buffer
                self.buffer[:-self.CHUNK_SIZE] = self.buffer[self.CHUNK_SIZE:]
                self.buffer[-self.CHUNK_SIZE:] = decoded_frame

                # apply the fourier transformation on the whole buffer (with zero-padding + hanning window)
                pad = np.pad(self.buffer * self.hanning_window, (0, len(self.buffer) * self.ZERO_PADDING), "constant")
                fft = np.fft.fft(pad)
                magnitude_data = abs(fft)

                # only use the first half of the fft output data
                magnitude_data = magnitude_data[:int(len(magnitude_data) / 2)]

                # HPS: multiply data by itself with different scalings (Harmonic Product Spectrum)
                magnitude_data_orig = copy.deepcopy(magnitude_data)
                for i in range(2, self.NUM_HPS + 1, 1):
                    hps_len = int(np.ceil(len(magnitude_data) / i))
                    magnitude_data[:hps_len] *= magnitude_data_orig[::i]  # multiply every i element

                # get the corresponding frequency array
                frequencies = np.fft.fftfreq(int((len(magnitude_data) * 2) / 1), 1. / self.SAMPLING_RATE)

                sound = self.note_detection(magnitude_data, frequencies)
                sound['execution_time'] = timedelta(seconds=time.perf_counter() - start_time).total_seconds()

                # display
                self.queue.put(sound)

            except Exception as e:
                sys.stderr.write('Error: Line {} {} {}\n'.format(sys.exc_info()[-1].tb_lineno, type(e).__name__, e))

    def stop(self):
        self.running = False
        self.stream.stop_stream()
        self.stream.close()
        self.audio_object.terminate()



"""
Copyright (C) 2025 drd <drd.ltt000@gmail.com>

This file is part of TunEd.

TunEd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TunEd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import random
import sys
from queue import Queue

from tuned.audio_analyzer import AudioAnalyzer


parser = argparse.ArgumentParser(prog='TunEd', description='Command line Tuner', epilog='')
parser.add_argument('--version', action='version', version='%(prog)s 0.2.1')
parser.add_argument('--verbose', '-v', action='count', default=0, help='Set verbosity level.')
parser.add_argument('--frequency', '-f', action='store', default=440, help='Set reference frequency.', type=int)
parser.add_argument('--display', '-d', action='extend', nargs='+', type=str, choices=['tuner', 'precision', 'frequency', 'signal_level', 'execution_time', 'egg'], help='Display detailed information.')

args = parser.parse_args()

VERBOSE = args.verbose
REF_FREQ = args.frequency
DISPLAY = args.display

default_display = ['tuner', 'precision', 'frequency', 'signal_level', 'execution_time']
to_display = DISPLAY if DISPLAY is not None else default_display

bold_char = '\033[1m'
default_color = '\033[0;0m'
color_dark_grey = '\033[38;2;67;70;75m'
color_green = '\033[38;2;0;255;0m'
color_red = '\033[38;2;255;0;0m'

gradients = {
        5: color_green,  # green
        10: '\033[38;2;63;192;0m',
        15: '\033[38;2;127;128;0m',
        20: '\033[38;2;192;63;0m',
        21: color_red  # red
    }

levels = ["▁", "▂", "▃", "▄", "▅", "▇", "█"]


def display_tuner(offset, note, octave):
    """
        ❱ ₋₄₅ │││││││││││┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃┃ G 1 │││││││││││││││││││││││││││││││││││││││││││││ ₊₄₅ ❰
    """
    abs_offset = abs(offset)

    if 0 <= abs_offset <= 5:
        color = gradients[5]
    elif 6 <= abs_offset <= 10:
        color = gradients[10]
    elif 11 <= abs_offset <= 15:
        color = gradients[15]
    elif 16 <= abs_offset <= 20:
        color = gradients[20]
    else:
        color = gradients[21]

    if abs_offset > 45:
        abs_offset = 45
    add = 45 - abs_offset
    left_offset = right_offset = 0
    right_add = left_add = 45
    l_arrow_color = l_max_color = r_max_color = r_arrow_color = color_dark_grey
    if offset < 0:
        left_offset = abs_offset
        right_offset = 0
        left_add = add
        right_add = 45
        l_arrow_color = color
        if offset >= 45:
            l_max_color = color
    elif offset > 0:
        left_offset = 0
        right_offset = abs_offset
        left_add = 45
        right_add = add
        r_arrow_color = color
        if offset <= -45:
            r_max_color = color

    l_arrow = f"{l_arrow_color}❱{default_color}"
    l_max = f"{l_max_color}₋₄₅{default_color}"

    l_offset = f"{color_dark_grey}{'│'*left_add:{left_add}}{color}{'┃'*left_offset:{left_offset}}{default_color}"
    r_offset = f"{color}{'┃'*right_offset:{right_offset}}{color_dark_grey}{'│'*right_add:{right_add}}{default_color}"

    c_note = f"{color}{note:2}{default_color}"
    c_octave = f"{color}{octave:1}{default_color}"

    r_max = f"{r_max_color}₊₄₅{default_color}"
    r_arrow = f"{r_arrow_color}❰{default_color}"

    d_tuner = f"{l_arrow:1} {l_max:3} {l_offset:45} {c_note:^2}{c_octave:>1} {r_offset:45} {r_max:3} {r_arrow:1}"

    return d_tuner


def display_output(sound):

    # datas
    magnitude = sound['magnitude']
    db = round(sound['magnitude_to_db'], 0)
    frequency = sound['frequency']
    phase = sound['phase']
    note = sound['note']
    octave = str(abs(sound['octave']))
    offset = sound['offset']
    execution_time = sound['execution_time']

    # infos
    d_tuner = f'{display_tuner(offset, note, octave):107}'
    d_precision = f'{offset:+5}¢'
    d_frequency = f'∿ {frequency:6}㎐'
    d_signal_level = f'{levels[int(abs(db//30))]} {db:6}㏈'
    d_phase = f'𝛗 {phase:3}㎭'
    d_exec = f'⧖ {execution_time:8}″'
    d_egg = f"{random.choice(list(gradients.values()))}{random.choice(list('🯅🯆🯇🯈')):^4}{default_color}"

    to_display_dict = {
        'tuner': d_tuner,
        'precision': d_precision,
        'frequency': d_frequency,
        'signal_level': d_signal_level,
        'execution_time': d_exec,
        'egg': d_egg
    }

    output = ""
    for d in to_display:
        output += f"[{to_display_dict[d]}]"

    return output


def tuned():
    try:
        q = Queue()
        aa = AudioAnalyzer(q, REF_FREQ)
        aa.start()
        print(f"{bold_char}@ {REF_FREQ}㎐{default_color}")

        while True:
            print("\r", display_output(q.get()), end='')

    except KeyboardInterrupt:
        print("\nExit.")
        aa.stop()
        aa.join()
        sys.exit()
